/************************************************************************
 *   IRC - Internet Relay Chat, m_core/burst.c
 *   This file is copyright (C) 2001 Andrew Suffield
 *                                    <asuffield@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "burst.h"
#include "client.h"
#include "s_user.h"
#include "struct.h"
#include "channel.h"
#include "ircd.h"
#include "send.h"
#include "s_serv.h"
#include "irc_string.h"
#include "user/user.h"

#include <string.h>

static char modebuf[BUFSIZE], parabuf[BUFSIZE], buf[BUFSIZE];

void
burst_to(struct Client *cptr)
{
  struct SLink *l;
  static char nickissent = 1;
  struct Channel *chptr;
  struct Client *acptr;

  /*
     ** Send it in the shortened format with the TS, if
     ** it's a TS server; walk the list of channels, sending
     ** all the nicks that haven't been sent yet for each
     ** channel, then send the channel itself -- it's less
     ** obvious than sending all nicks first, but on the
     ** receiving side memory will be allocated more nicely
     ** saving a few seconds in the handling of a split
     ** -orabidoo
   */

  nickissent = 3 - nickissent;

  for (chptr = channel; chptr; chptr = chptr->nextch)
    {
      for (l = chptr->members; l; l = l->next)
	{
	  acptr = l->value.cptr;
	  if (acptr->nicksent != nickissent)
	    {
	      acptr->nicksent = nickissent;
	      if (acptr->from != cptr)
		sendnick_TS(cptr, acptr);
	    }
	}
      /* Don't burst an empty channel's modes, it's been created on the other side */
      /* Also don't bother bursting if I have this channel locally locked (and therefore
       * cleared), since I'm waiting for the remote server to send me that channel image,
       * but do call send_channel_users() to join all local users to the remote channel
       */
      if (chptr->users != 0)
	{
	  if ((chptr->remotets == -1) || (chptr->channelts < chptr->remotets))
	    {
	      send_channel_modes(cptr, chptr);
#ifdef TOPIC_INFO
	      if (chptr->topic_time)
		sendto_one(cptr, ":%s STOPIC %s %s %lu :%s",
			   me.name, chptr->chname, chptr->topic_nick,
			   chptr->topic_time, chptr->topic);
#endif
	    }
	  else
	    {
	      send_channel_users(cptr, chptr);
	    }
	}
    }

  for (acptr = &me; acptr; acptr = acptr->prev)
    if (acptr->nicksent != nickissent)
      {
	acptr->nicksent = nickissent;
	if (acptr->from != cptr)
	  sendnick_TS(cptr, acptr);
      }

  sendto_one(cptr, ":%s BURSTDONE", me.name);
}

/*
 * Here's the stuff to send a channel over to the other side
 * We've already agreed which version of the channel we're going
 *  to use, and it's this one. The other side will clear out
 *  it's modes for us.
 */

void
send_channel_modes(struct Client *cptr, struct Channel *chptr)
{
  Link *l, *anop = NULL, *skip = NULL;
  int n = 0;
  char *t;
  int i;

  if (*chptr->chname != '#')
    return;

  *modebuf = *parabuf = '\0';
  channel_modes(cptr, modebuf, parabuf, chptr, 1);

  if (*parabuf)
    strcat(parabuf, " ");
  ircsprintf(buf, ":%s SJOIN %lu %s %s %s:", me.name,
	     chptr->channelts, chptr->chname, modebuf, parabuf);
  t = buf + strlen(buf);
  for (l = chptr->members; l && l->value.cptr; l = l->next)
    if (l->flags & CUFLAG_CHANOP)
      {
	anop = l;
	break;
      }
  /* follow the channel, but doing anop first if it's defined
     **  -orabidoo
   */
  l = NULL;
  for (;;)
    {
      if (anop)
	{
	  l = skip = anop;
	  anop = NULL;
	}
      else
	{
	  if (l == NULL || l == skip)
	    l = chptr->members;
	  else
	    l = l->next;
	  if (l && l == skip)
	    l = l->next;
	  if (l == NULL)
	    break;
	}
      if (l->flags & CUFLAG_CHANOP)
	*t++ = '@';
      if (l->flags & CUFLAG_VOICE)
	*t++ = '+';
      strcpy(t, l->value.cptr->name);
      t += strlen(t);
      *t++ = ' ';
      n++;
      if (t - buf > BUFSIZE - 80)
	{
	  *t++ = '\0';
	  if (t[-1] == ' ')
	    t[-1] = '\0';
	  sendto_one(cptr, "%s", buf);
	  ircsprintf(buf, ":%s SJOIN %lu %s 0 :",
		     me.name, chptr->channelts, chptr->chname);
	  t = buf + strlen(buf);
	  n = 0;
	}
    }

  if (n)
    {
      *t++ = '\0';
      if (t[-1] == ' ')
	t[-1] = '\0';
      sendto_one(cptr, "%s", buf);
    }

  /* Do all the mask types now */
  for (i = 0; cmodes[i].value; i++)
    if (cmodes[i].type == CMTYPE_MASK)
      {
	/* This should be rewritten to build multiple mode changes into one message */
	struct SLink *cmlink;
	for (cmlink = chptr->mode_mask_list[cmodes[i].value]; cmlink;
	     cmlink = cmlink->next)
	  {
	    struct channel_mask *mask = cmlink->value.cmptr;
	    if (mask)
	      sendto_one(cptr, ":%s MODE %s +%c %s",
			 me.name, chptr->chname,
			 cmodes[i].letter, build_channel_mask(mask));
	  }
      }
}

/* Clear out all the local channel modes in preparation for
 * receiving a new list. Leave the flag modes alone, m_sjoin will deal
 * with those. Clear all mask and user flags, we're getting a new
 * set across
 *
 * NOTE: This does not propagate to other servers, it only clears
 * locally
 */
void
clear_channel_modes(struct Client *sptr, struct Channel *chptr)
{
  int what = 0;
  Link *l;
  char *mbuf = modebuf;
  int pargs = 0;
#ifdef SERVERHIDE
  struct Client *from = IsServer(sptr) ? &me : sptr;
#else
  struct Client *from = sptr;
#endif
  int i;

  /* First, channel masks */
  for (i = 0; cmodes[i].value; i++)
    if (cmodes[i].type == CMTYPE_MASK)
      {
	/* This should be rewritten to build multiple mode changes into one message */
	struct SLink *cmlink;
	for (cmlink = chptr->mode_mask_list[cmodes[i].value]; cmlink;)
	  {
	    struct channel_mask *mask = cmlink->value.cmptr;

	    if (mask)
	      sendto_channel_butserv(chptr, &me, ":%s MODE %s -%c %s!%s@%s",
				     from->name, chptr->chname,
				     cmodes[i].letter,
				     mask->nick, mask->user, mask->host);
	    delete_channel_mask(chptr, cmlink, cmodes[i].value);
	    /* Taking advantage of how delete_channel_mask() works */
	    if (!mask)
	      cmlink = cmlink->next;
	  }
      }

  /* Now do the user IDs */
  *mbuf = parabuf[0] = '\0';

  for (l = chptr->members; l && l->value.cptr; l = l->next)
    {
      if (l->flags & CUFLAG_CHANOP)
	{
	  if (what != -1)
	    {
	      *mbuf++ = '-';
	      what = -1;
	    }
	  *mbuf++ = 'o';
	  strcat(parabuf, l->value.cptr->name);
	  strcat(parabuf, " ");
	  pargs++;
	  if (pargs >= 4)
	    {
	      *mbuf = '\0';
	      sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
				     from->name, chptr->chname, modebuf,
				     parabuf);
	      mbuf = modebuf;
	      *mbuf = parabuf[0] = '\0';
	      pargs = what = 0;
	    }
	  l->flags &= ~CUFLAG_CHANOP;
	}
      if (l->flags & CUFLAG_VOICE)
	{
	  if (what != -1)
	    {
	      *mbuf++ = '-';
	      what = -1;
	    }
	  *mbuf++ = 'v';
	  strcat(parabuf, l->value.cptr->name);
	  strcat(parabuf, " ");
	  pargs++;
	  if (pargs >= 4)
	    {
	      *mbuf = '\0';
	      sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s",
				     from->name, chptr->chname, modebuf,
				     parabuf);
	      mbuf = modebuf;
	      *mbuf = parabuf[0] = '\0';
	      pargs = what = 0;
	    }
	  l->flags &= ~CUFLAG_VOICE;
	}
    }
  if (mbuf != modebuf)
    {
      *mbuf = '\0';
      sendto_channel_butserv(chptr, sptr, ":%s MODE %s %s %s", from->name,
			     chptr->chname, modebuf, parabuf);
    }
}

void
sendnick_TS(struct Client *cptr, struct Client *acptr)
{
  static char ubuf[BUFSIZE];

  if (IsPerson(acptr))
    {
      send_umode(NULL, acptr, 0, ubuf);
      if (!*ubuf)
	{			/* trivial optimization - Dianora */

	  ubuf[0] = '+';
	  ubuf[1] = '\0';
	  /*    original was    strcpy(ubuf, "+"); */
	}

      sendto_one(cptr, "NICK %s %d %lu %s %s %s %s %lu :%s", acptr->name,
		 acptr->hopcount + 1, acptr->tsinfo, ubuf,
		 acptr->username, acptr->host,
		 acptr->user->server, (long int) acptr->ip.s_addr,
		 acptr->info);
      /* Let's propogate AWAY status in the burst as well */
      if (acptr->user->away)
	sendto_one(cptr, ":%s AWAY :%s", acptr->name, acptr->user->away);
    }
}

void
send_channel_users(struct Client *cptr, struct Channel *chptr)
{
  Link *l, *anop = NULL, *skip = NULL;
  int n = 0;
  char *t;

  if (*chptr->chname != '#')
    return;

  if (*parabuf)
    strcat(parabuf, " ");
  ircsprintf(buf, ":%s SJOIN %lu %s + :", me.name,
	     chptr->channelts, chptr->chname);
  t = buf + strlen(buf);
  for (l = chptr->members; l && l->value.cptr; l = l->next)
    if (l->flags & CUFLAG_CHANOP)
      {
	anop = l;
	break;
      }
  /* follow the channel, but doing anop first if it's defined
     **  -orabidoo
   */
  l = NULL;
  for (;;)
    {
      if (anop)
	{
	  l = skip = anop;
	  anop = NULL;
	}
      else
	{
	  if (l == NULL || l == skip)
	    l = chptr->members;
	  else
	    l = l->next;
	  if (l && l == skip)
	    l = l->next;
	  if (l == NULL)
	    break;
	}
      if (l->flags & CUFLAG_CHANOP)
	*t++ = '@';
      if (l->flags & CUFLAG_VOICE)
	*t++ = '+';
      strcpy(t, l->value.cptr->name);
      t += strlen(t);
      *t++ = ' ';
      n++;
      if (t - buf > BUFSIZE - 80)
	{
	  *t++ = '\0';
	  if (t[-1] == ' ')
	    t[-1] = '\0';
	  sendto_one(cptr, "%s", buf);
	  ircsprintf(buf, ":%s SJOIN %lu %s 0 :",
		     me.name, chptr->channelts, chptr->chname);
	  t = buf + strlen(buf);
	  n = 0;
	}
    }

  if (n)
    {
      *t++ = '\0';
      if (t[-1] == ' ')
	t[-1] = '\0';
      sendto_one(cptr, "%s", buf);
    }
}
